@@ -168,8 +168,34 @@ int __pthread_mutexattr_getkind_np(const pthread_mutexattr_t *attr, int *kind)
 }
 weak_alias (__pthread_mutexattr_getkind_np, pthread_mutexattr_getkind_np)
 
-int pthread_once(pthread_once_t * once_control, void (*init_routine)(void))
+/* Once-only execution */
+
+static pthread_mutex_t once_masterlock = PTHREAD_MUTEX_INITIALIZER;
+static pthread_cond_t once_finished = PTHREAD_COND_INITIALIZER;
+
+enum { NEVER = 0, IN_PROGRESS = 1, DONE = 2 };
+
+int __pthread_once(pthread_once_t * once_control, void (*init_routine)(void))
 {
-  if (testandset(once_control) == 0) init_routine();
+  /* Test without locking first for speed */
+  if (*once_control == DONE) return 0;
+  /* Lock and test again */
+  pthread_mutex_lock(&once_masterlock);
+  /* If init_routine is being called from another routine, wait until
+     it completes. */
+  while (*once_control == IN_PROGRESS) {
+    pthread_cond_wait(&once_finished, &once_masterlock);
+  }
+  /* Here *once_control is stable and either NEVER or DONE. */
+  if (*once_control == NEVER) {
+    *once_control = IN_PROGRESS;
+    pthread_mutex_unlock(&once_masterlock);
+    init_routine();
+    pthread_mutex_lock(&once_masterlock);
+    *once_control = DONE;
+    pthread_cond_broadcast(&once_finished);
+  }
+  pthread_mutex_unlock(&once_masterlock);
   return 0;
 }
+weak_alias (__pthread_once, pthread_once)
