@@ -75,7 +75,7 @@ int pthread_key_delete(pthread_key_t key)
 int __pthread_setspecific(pthread_key_t key, const void * pointer)
 {
   pthread_descr self = thread_self();
-  if (key >= PTHREAD_KEYS_MAX) return EINVAL;
+  if (key >= PTHREAD_KEYS_MAX || !pthread_keys[key].in_use) return EINVAL;
   self->p_specific[key] = (void *) pointer;
   return 0;
 }
@@ -86,7 +86,7 @@ weak_alias (__pthread_setspecific, pthread_setspecific)
 void * __pthread_getspecific(pthread_key_t key)
 {
   pthread_descr self = thread_self();
-  if (key >= PTHREAD_KEYS_MAX)
+  if (key >= PTHREAD_KEYS_MAX || !pthread_keys[key].in_use)
     return NULL;
   else
     return self->p_specific[key];
@@ -97,14 +97,29 @@ weak_alias (__pthread_getspecific, pthread_getspecific)
 
 void __pthread_destroy_specifics()
 {
-  int i;
+  unsigned int round;
   pthread_descr self = thread_self();
-  destr_function destr;
-  void * data;
 
-  for (i = 0; i < PTHREAD_KEYS_MAX; i++) {
-    destr = pthread_keys[i].destr;
-    data = self->p_specific[i];
-    if (destr != NULL && data != NULL) destr(data);
-  }
+  for (round = 0; round < PTHREAD_DESTRUCTOR_ITERATIONS; ++round)
+    {
+      int found_nonzero = 0;
+      unsigned int i;
+      destr_function destr;
+      void * data;
+
+      for (i = 0; i < PTHREAD_KEYS_MAX; ++i)
+	{
+	  destr = pthread_keys[i].destr;
+	  data = self->p_specific[i];
+	  if (destr != NULL && data != NULL)
+	    {
+	      destr (data);
+	      found_nonzero |= self->p_specific[i] != NULL;
+	    }
+	}
+
+      if (found_nonzero == 0)
+	/* No more non-NULL data with destructor.  */
+	break;
+    }
 }
