@@ -17,7 +17,7 @@
 
 static inline void acquire(int * spinlock)
 {
-  while (testandset(spinlock)) sched_yield();
+  while (testandset(spinlock)) __sched_yield();
 }
 
 static inline void release(int * spinlock)
