@@ -19,16 +19,11 @@
 
 #include <pthread.h>
 
-#ifdef USE_IN_LIBIO
 typedef pthread_mutex_t _IO_lock_t;
 
 /* We need recursive (counting) mutexes.  */
 #define _IO_lock_initializer PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP
 
-#else
-
-#endif
-
 
 #define _IO_cleanup_region_start(_fct, _fp) \
      __libc_cleanup_region_start (_fct, _fp)
