@@ -18,6 +18,7 @@
    Boston, MA 02111-1307, USA.  */
 
 #include <errno.h>
+#include <limits.h>
 #include <stdlib.h>
 
 /* This is from <pthread.h>.  */
@@ -59,28 +60,10 @@ weak_alias (__pthread_return_0, pthread_cond_wait)
 weak_alias (__pthread_return_0, pthread_cond_signal)
 weak_alias (__pthread_return_0, pthread_cond_broadcast)
 
-#if 0
-/* I don't think it is a good idea to define these functions.  Returning
-   a positive result might make the caller think the wanted action was
-   carried out but in fact it was not.  --drepper@gnu  */
-weak_alias (__pthread_return_0, pthread_create)
-weak_alias (__pthread_return_0, pthread_join)
-weak_alias (__pthread_return_0, pthread_detach)
-weak_alias (__pthread_return_0, pthread_cond_timedwait)
-weak_alias (__pthread_return_0, pthread_key_delete)
-weak_alias (__pthread_return_0, pthread_once)
-weak_alias (__pthread_return_0, pthread_cancel)
-
-/* Those are pthread functions which return void. */
-weak_alias (__pthread_return_void, pthread_testcancel)
-#endif
 
 /* Those are pthread functions which return 1 if successful. */
 weak_alias (__pthread_return_1, pthread_equal)
 
-/* Those are pthread functions which return != 0 if unsuccessful. */
-weak_alias (__pthread_return_1, __pthread_key_create)
-
 /* pthread_exit () is a special case. */
 void
 weak_function
@@ -105,3 +88,74 @@ void
 __pthread_return_void (void)
 {
 }
+
+
+/* The key and thread-specific data functions cannot simply be dummies
+   since data might be stored in any case.  */
+
+/* Number of supported keys.  */
+
+/* Table of keys. */
+struct pthread_key_struct
+  {
+    int occupied;
+    void (*destr) (void *);
+    const void *data;
+  };
+
+static struct pthread_key_struct keys[PTHREAD_KEYS_MAX] =
+  {
+    { 0, NULL, NULL },
+  };
+
+
+int
+__attribute__ ((weak))
+__pthread_key_create (pthread_key_t *key, void (*destructor) (void *))
+{
+  unsigned int cnt;
+
+  for (cnt = 0; cnt < PTHREAD_KEYS_MAX; ++cnt)
+    if (! keys[cnt].occupied)
+      {
+	keys[cnt].occupied = 0;
+	keys[cnt].destr = destructor;
+	*key = cnt;
+	return 0;
+      }
+  return EAGAIN;
+}
+
+
+int
+__attribute__ ((weak))
+__pthread_key_delete (pthread_key_t key)
+{
+  if (key >= PTHREAD_KEYS_MAX || ! keys[key].occupied)
+    return EINVAL;
+  keys[key].occupied = 0;
+  keys[key].destr = NULL;
+  keys[key].data = NULL;
+  return 0;
+}
+
+
+int
+__attribute__ ((weak))
+__pthread_setspecific (pthread_key_t key, const void *ptr)
+{
+  if (key >= PTHREAD_KEYS_MAX || ! keys[key].occupied)
+    return EINVAL;
+  keys[key].data = ptr;
+  return 0;
+}
+
+
+void *
+__attribute__ ((weak))
+__pthread_getspecific (pthread_key_t key)
+{
+  if (key >= PTHREAD_KEYS_MAX || ! keys[key].occupied)
+    return NULL;
+  return (void *) keys[key].data;
+}
